#
# TWiki WikiClone ($wikiversion has version info)
#
# Copyright (C) 2000-2001 Andrea Sterbini, a.sterbini@flashnet.it
# Copyright (C) 2001 Peter Thoeny, Peter@Thoeny.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, published at 
# http://www.gnu.org/copyleft/gpl.html
#
# =========================
#
# This is an empty TWiki plugin. Use it as a template
# for your own plugins; see TWiki.TWikiPlugins for details.
#
# Each plugin is a package that contains the subs:
#
#   initPlugin           ( $topic, $web, $user, $installWeb )
#   commonTagsHandler    ( $text, $topic, $web )
#   startRenderingHandler( $text, $web )
#   outsidePREHandler    ( $text )
#   insidePREHandler     ( $text )
#   endRenderingHandler  ( $text )
#
# initPlugin is required, all other are optional. 
# For increased performance, all handlers except initPlugin are
# disabled. To enable a handler remove the leading DISABLE_ from
# the function name.
# 
# NOTE: To interact with TWiki use the official TWiki functions
# in the &TWiki::Func module. Do not reference any functions or
# variables elsewhere in TWiki!!


# =========================
package TWiki::Plugins::EvEmPlugin;     # change the package name!!!

# =========================
use vars qw(
        $web $topic $user $installWeb $VERSION $debug
        $exampleCfgVar $scriptUrlPath $raw $pubDir 
    );
$raw = "";
$pubDir = "";
$VERSION = '1.000';

# =========================
sub initPlugin
{
    ( $topic, $web, $user, $installWeb ) = @_;

    # check for Plugins.pm versions
    if( $TWiki::Plugins::VERSION < 1 ) {
        &TWiki::Func::writeWarning( "Version mismatch between EvEmPlugin and Plugins.pm" );
        return 0;
    }

    # Get plugin preferences, the variable defined by:          * Set EXAMPLE = ...
    $exampleCfgVar = &TWiki::Prefs::getPreferencesValue( "EVEMPLUGIN_EXAMPLE" ) || "default";

    # Get plugin debug flag
    $debug = &TWiki::Func::getPreferencesFlag( "EVEMPLUGIN_DEBUG" );

    # Setup EvEm Defaults:
    # Get raw topic text for use in plugin

    $raw = &TWiki::Store::readTopicRaw( $web, $topic );    
    $pubDir = &TWiki::Func::getPubDir();

    # Plugin correctly initialized
    &TWiki::Func::writeDebug( "- TWiki::Plugins::EvEmPlugin::initPlugin( $web.$topic ) is OK" ) if $debug;
    return 1;
}

# ========================= #############################
# = Custom EvEm Bit ======= #############################
# ========================= #############################

# Utils
#########################################################

sub extractMetaData
{

    my ( $themetaitem, $themetafield, $valuepair, $theweb, $thetopic ) = @_;
    
    if ( $themetafield )
    {
        $themetafield = "name=\"$themetafield\" ";
    }
    $temp = "FIELD";

    if ( $thetopic ne $topic )
    {
        my $oldraw = $raw;
        $raw = &TWiki::Store::readTopicRaw( $theweb, $thetopic );
    }

        $raw =~ /%META\:$themetaitem\{$themetafield(.+?)}/i;
        my $metavalue = $1;
        $metavalue = &TWiki::Func::extractNameValuePair( $metavalue, $valuepair );
        if ( $oldraw ) { $raw = $oldraw }; # put the default raw value back
    
    return $metavalue;


}

sub renderTextAndVariables
{

    my ( $theText ) = @_;
    $theText = &TWiki::Func::expandCommonVariables( $theText );
    return &TWiki::Func::renderText( $theText );
}

sub extractFormName
{

    my $title = extractMetaData( "FORM", "", "name", $web, $topic );
    $title =~ s/^.*?\.//g;
    return $title;
}

sub getRandomNumber
{
    return int(rand()*10000);
}


# Renderers
#########################################################

sub handleKeywords
{
    my @keywordsarray;
    my @safekeywordsarray;
    my $linkreturn;
    my ( $attributes ) = @_;

    my $alttopic = &TWiki::Func::extractNameValuePair( $attributes, "topic" ); 
    my $keywords = &TWiki::Func::extractNameValuePair( $attributes, "keywords" ); 
    my $base = &TWiki::Func::extractNameValuePair( $attributes, "base" );
    my $sep = &TWiki::Func::extractNameValuePair( $attributes, "separator" );

    if ( ! $base ) { $base = "SectionIndex" };
    if ( ! $sep ) { $sep = ":" };
    if ( ! $alttopic ) { $alttopic = $topic };

    if ( ! $keywords ) {
        $keywords = extractMetaData( "FIELD", "ArticleSection", "value", $web, $alttopic );
    }

    @keywordsarray = split(/,/,$keywords);
    @safekeywordsarray = split(/,/,$keywords);

    foreach $val (@safekeywordsarray)
    {
        $val =~ s/^\s+|\s+$//g;
        $val =~ s/\s/%20/g;
    }

    foreach $val (@keywordsarray)
    {
        $val =~ s/^\s+|\s+$//g;
    }

    $linkreturn .= "<a href=\"$base?count=20&index=$safekeywordsarray[0]&skin=home\">$keywordsarray[0]</a>";
    if ( $keywordsarray[1] )
    {
        $linkreturn .= " $sep <a href=\"$base?count=20&index=$safekeywordsarray[0],%20$safekeywordsarray[1]&skin=home\">$keywordsarray[1]</a>";
    }
    if ( $keywordsarray[2] )
    {
        $linkreturn .= " $sep <a href=\"$base?count=20&index=$safekeywordsarray[0],%20$safekeywordsarray[1],%20$safekeywordsarray[2]&skin=home\">$keywordsarray[2]</a>";
    }

    return $linkreturn;

}


sub renderMetaField
{
    my ( $attributes ) = @_;

    my $fieldname = &TWiki::Func::extractNameValuePair( $attributes, "field" );
    my $alttopic = &TWiki::Func::extractNameValuePair( $attributes, "topic" );
    my $dorendertext = &TWiki::Func::extractNameValuePair( $attributes, "dorendertext" );
    
    if ( ! $alttopic ) { $alttopic = $topic };
    if ( ! $fieldname ) { $fieldname = $attributes };

    $fieldname =~ s/\"//g;

    my $fieldvalue = extractMetaData( "FIELD", $fieldname, "value", $web, $alttopic );

    $fieldvalue =~ s/%_N_%/&nbsp;<br \/>/go;
    $fieldvalue =~ s/%_Q_%/"/go;

    if ( $dorendertext )
    {
        $fieldvalue = &TWiki::Func::expandCommonVariables( $fieldvalue );
        $fieldvalue = &TWiki::Func::renderText( $fieldvalue );
    }

    return $fieldvalue;
}

sub renderTitle
{
    my $title = extractMetaData( "FIELD", "Title", "value", $web, $topic );
    return $title;
}

sub renderKeywordSection
{
    # Pull in any variables
    my ( $section )= @_;

    # If none passed by variable, check form fields
    if (! $section ) {
        $section = extractMetaData( "FIELD", "ArticleSection", "value", $web, $topic );
    }

    # If no form field, check querystrings.
    if (! $section ) { 
        $section = &TWiki::handleUrlParam("index");
    }

    my @section = split(/,/,$section);
    $section = $section[0];

    # Screw it, can't find anything just return 'Home'
    if (! $section || $section eq " " || $section eq "\*" ) { $section = "Home" }

    return $section;
}

sub renderSimilarKeywordLinks
{

    my $linktext;
    my @keywords;
    my @keywordsSafe;
    my $keylist;
    my $safekeylist;
    my ( $attributes ) = @_;

    my $header = &TWiki::Func::extractNameValuePair( $attributes, "header" );
    my $footer = &TWiki::Func::extractNameValuePair( $attributes, "footer" );
    my $base = &TWiki::Func::extractNameValuePair( $attributes, "base" );
    my $keywords = &TWiki::Func::extractNameValuePair( $attributes, "keywords" );
    my $sep = &TWiki::Func::extractNameValuePair( $attributes, "separator" );
    my $joiner = &TWiki::Func::extractNameValuePair( $attributes, "joiner" );

    if ( ! $keywords )
    {
        $keywords = extractMetaData( "FIELD", "ArticleKeywords", "value", $web, $topic );
    }
   
    if ( $keywords )
    {
        $linktext .= $header;
        if (! $base ) { $base = "KeywordIndex" };
        if (! $sep ) { $sep = ":" };
        if (! $joiner ) { $joiner = "+" };
        
        @keywords = split(/,/,$keywords);
        @safekeywords = split(/,/,$keywords);

        foreach $val (@safekeywords)
        {
            $val =~ s/^\s+|\s+$//g;
            $val =~ s/\s/%20/g;
        }

        foreach $val (@keywords)
        {
            $val =~ s/^\s+|\s+$//g;
        }

        for ( my $i = 0; $i < @keywords; $i++ )
        {
            if ($keywords[$i])
            {
                $keylist .= "$keywords[$i] $joiner ";
                $safekeylist .= "$safekeywords[$i],%20";
                $linktext .= "<a href=\"$base\?count=20&index=$safekeywords[$i]&skin=home\">$keywords[$i]</a> $sep "
            }
        }
        
        my $tempjoinerval = " $joiner ";
        $keylist =~ s/$tempjoinerval$//g;
        $safekeylist =~ s/,%20$//g;

        if ( $keywords[1] )
        {
            $linktext .= "<a href=\"$base?count=20&index=$safekeywords[0],%20$safekeywords[1]&skin=home\">$keywords[0] $joiner $keywords[1]</a>";
        }
        if ( $keywords[2] )
        {
            $linktext .= " $sep <a href=\"$base?count=20&index=$safekeywords[0],%20$safekeywords[1],%20$safekeywords[2]&skin=home\">$keywords[0] $joiner $keywords[1] $joiner $keywords[2]</a>";
        }
        if ( $keywords[3] )
        {
            $linktext .= " $sep <a href=\"$base?count=20&index=$safekeylist&skin=home\">$keylist</a>";
        }
        $linktext .= $footer;
    }


    return $linktext;

    }

sub renderImage {

    my $altweb;
    my $alttopic;
    my $imagename;
    my $align;
    my $border;
    my $clear;
    my $height;
    my $width;

    my ( $attributes ) = @_;

    $altweb = &TWiki::Func::extractNameValuePair( $attributes, "web" );
    $alttopic = &TWiki::Func::extractNameValuePair( $attributes, "topic" );
    $imagename = &TWiki::Func::extractNameValuePair( $attributes, "image" );
    $align = &TWiki::Func::extractNameValuePair( $attributes, "align" );
    $border = &TWiki::Func::extractNameValuePair( $attributes, "border" );
    $clear = &TWiki::Func::extractNameValuePair( $attributes, "clear" );
    $height = &TWiki::Func::extractNameValuePair( $attributes, "height" );
    $width = &TWiki::Func::extractNameValuePair( $attributes, "width" );

    if ( ! $altweb ) { $altweb = $web };
    if ( ! $alttopic ) { $alttopic = $topic };
    if (! $align ) { $align = "right" };
    if (! $border ) { $border = "0" };
    if (! $clear ) {
        $clear = "";
    } else {
        $clear = " clear=\"$clear\"";
    }
    if (! $height ) {
        $height = "";
    } else {
        $height = " height=\"$height\"";
    }
    if (! $width ) {
        $width = "";
    } else {
        $width = " width=\"$width\"";
    }

    if ( $imagename =~ /(jpg|gif|png)$/i ) {
        return "<img src=\"$imagename\" align=\"$align\" border=\"$border\"$height$width$clear />";
    } else {
        return "";    #--- Nothing yet
    }

}

sub renderThread {

    my ( $attributes ) = @_;
    my $thread;
    my $padding;
    my @depth;
    my $depth;

    $thread = &TWiki::Func::extractNameValuePair( $attributes, "thread" );
    $padding = &TWiki::Func::extractNameValuePair( $attributes, "padding" );
    $delimiter = &TWiki::Func::extractNameValuePair( $attributes, "delimiter" );

    @depth = split( /$delimiter/, $thread );

    foreach $level (@depth) {
        $depth .= "+";
    }

    $depth =~ s/\+\+//;
    $depth =~ s/\+/$padding/g;

    return $depth;

}

sub renderForm {

    my $return;
    my $value;
    my $linkval;
    my @vals;
    my ( $attributes ) = @_;

    #--- Extract name/value pairs and assign to variables

    my $header       = &TWiki::Func::extractNameValuePair( $attributes, "header" );
    my $footer       = &TWiki::Func::extractNameValuePair( $attributes, "footer" );

    my $web          = &TWiki::Func::extractNameValuePair( $attributes, "web" );
    my $newtopicname = &TWiki::Func::extractNameValuePair( $attributes, "newtopicname" );
    my $thread       = &TWiki::Func::extractNameValuePair( $attributes, "thread" );
    my $threadid     = &TWiki::Func::extractNameValuePair( $attributes, "threadid" );
    my $formtemplate = &TWiki::Func::extractNameValuePair( $attributes, "formtemplate" );
    my $formaction   = &TWiki::Func::extractNameValuePair( $attributes, "action" );
    my $parent       = &TWiki::Func::extractNameValuePair( $attributes, "parent" );
    my $buttontext   = &TWiki::Func::extractNameValuePair( $attributes, "button" );
    my $class        = &TWiki::Func::extractNameValuePair( $attributes, "class" );
    my $displayon    = &TWiki::Func::extractNameValuePair( $attributes, "displayon" );

    if ( $displayon ) {
        $currentform = &extractFormName;
        if ( $displayon ne $currentform ) { return "" }
    }

    if (! $buttontext ) { $buttontext = "Submit" };
    if (! $class ) { $class = "boardform" };
    if ( $formaction ne "save" ) { $formaction = "preview" }; # we only want save or preview
 

    #--- Read in templatetopic and split into lines
    
    my $rawform = &TWiki::Store::readTopicRaw( $web, $formtemplate );
    my @lines =  split( /\n/, $rawform );
    foreach $line ( @lines )
    {
        if ( $line =~ /^\s*\|/ )
        {
            $line =~ s/^\s*\|//o ; # Remove first | on line
            my ( $name, $type, $size, $vals, $tooltip ) = split( /\|/, $line );

            $name =~ s/^\s+|\s+$//g; #--- Remove leading and trailing spaces
            $type =~ s/^\s+|\s+$//g; #--- Remove leading and trailing spaces
            $size =~ s/^\s+|\s+$//g; #--- Remove leading and trailing spaces
            $vals =~ s/^\s+|\s+$//g; #--- Remove leading and trailing spaces
            $tooltip =~ s/^\s+|\s+$//g; #--- Remove leading and trailing spaces
            
            #--- Render field
            
            if ( $line =~ /\*Name\*/ ) {

                #--- Its the first header line. Ignore.

            } elsif ( $name eq "Thread" || $name eq "Threadid"   ) {
    
                #--- Ignore special board fields. #### yes this is a hack. Needs  to be changed.

    
            } elsif ( $type eq "text" ) {
    
                $value .= "<tr><td>$name:</td><td><input class=\"$class\" name=\"$name\" size=\"$size\" type=\"input\" value=\"$vals\" /></td></tr>\n";
    
            } elsif ( $type eq "textarea" ) {
    
                my $cols = 30;
                my $rows = 5;
                $value .= "<tr><td>$name:</td><td><textarea class=\"$class\" cols=\"$cols\" rows=\"$rows\" name=\"$name\">$vals</textarea></td></tr>\n";
    
            } elsif ( $type eq "select" ) {
                
                $linkval = "";
                @vals = split(/,/,$vals);    
                foreach $val (@vals) {
                    $val =~ s/^\s+|\s+$//g; #--- Remove leading and trailing spaces
                    $linkval .= "<option class=\"$class\" name=\"$val\">$val</a>\n";
                }
                $value .= "<tr><td>$name:</td><td><select name=\"$name\">\n";
                $value .= $linkval;
                $value .= "</select></td></tr>\n";
            
            } elsif ( $type eq "radio" ) {
                
                $linkval = "";
                @vals = split(/,/,$vals);    
                foreach $val (@vals) {
                    $val =~ s/^\s+|\s+$//g; #--- Remove leading and trailing spaces
                    $linkval .= "<input type=\"radio\" name=\"$name\" value=\"$val\" class=\"$class\">\n";
                }
                $value .= "<tr><td>$name:</td><td>\n";
                $value .= $linkval;
                $value .= "</td></tr>\n";
    
            } else {
    
                $value .= "<tr><td>$name:</td><td><input class=\"$class\" name=\"$name\" size=\"$size\" type=\"$type\" value=\"$vals\" /></td></tr>\n";
    
            }
        }
    }

    $return .= $header;
    $return .= "<form name=\"main\" action=\"/bin/$formaction/$web/\">\n";
    $return .= "<input type=\"hidden\" name=\"topic\" value=\"$newtopicname\" />\n";
    if ( $thread ) { $return .= "<input type=\"hidden\" name=\"Thread\" value=\"$thread\" />\n" };
    if ( $threadid ) { $return .= "<input type=\"hidden\" name=\"Threadid\" value=\"$threadid\" />\n" };
    $return .= "<input type=\"hidden\" name=\"formtemplate\" value=\"$formtemplate\" />\n";
    $return .= "<input type=\"hidden\" name=\"topicparent\" value=\"$parent\" />\n";
    $return .= "<input type=\"hidden\" name=\"cmd\" value=\"\" />\n";
    $return .= "<table class=\"$class\">\n";
    $return .= $value;
    $return .= "<tr valign=\"top\"><td>Comment:</td><td><textarea class=\"$class\" name=\"text\" wrap=\"virtual\" rows=\"5\" cols=\"40\"></textarea></td></tr>\n"; 
    $return .= "<tr><td>&nbsp;</td><td><input class=\"$class\" type=\"submit\" value=\" $buttontext \" /></td></tr>\n";
    $return .= "</table>\n";
    $return .= "</form>\n";
    $return .= $footer;

}

sub renderAuthorUserName
{
    my ( $alttopic ) = @_;
    $alttopic = &TWiki::Func::extractNameValuePair( $alttopic );
    my ($revdate, $revuser, $maxrev) = &TWiki::Func::getRevisionInfo( $web, $alttopic );
    return &TWiki::Func::renderText( $revuser );
}

sub handleEnvVariable
{   
    my( $theVar ) = @_;
    my $value = $ENV{$theVar} || "";
    return $value;
}

sub renderFormData {

    my $return;
    my $val;
    my @vals;
    my ( $attributes ) = @_;

    #--- Extract name/value pairs and assign to variables

    my $header       = &TWiki::Func::extractNameValuePair( $attributes, "header" );
    my $footer       = &TWiki::Func::extractNameValuePair( $attributes, "footer" );

    my $displayon    = &TWiki::Func::extractNameValuePair( $attributes, "displayon" );
    my $displayfields    = &TWiki::Func::extractNameValuePair( $attributes, "fields" );
    
    if ( $displayon )
    {
        my $currentform = &extractFormName;
        if ( $displayon ne $currentform ) { return "" }
    }

    if ( $header )
    {
        $return .= $header;
    }
    else
    {
        $return .= "<table>";
    }

    @vals = split ( /,/,$displayfields );
    foreach $val (@vals)
    {
        $return .= "<tr><td><b>" . extractMetaData( "FIELD", $val, "title", $web, $topic ) . "</b>:</td><td>" . extractMetaData( "FIELD", $val, "value", $web, $topic ) . "</td></tr>";
    }
    
     if ( $footer )
    {
        $return .= $footer;
    }
    else
    {
        $return .= "</table>";
    } 

    return $return;
}

sub renderDocumentID
{
    return getRandomNumber();
}

sub renderResponseParentLink {

    my ( $return, $thread, $parenttopic, $parenttitle, $tempjoinerval, @threadvals);
    my ( $attributes ) = @_;

    #--- Extract name/value pairs and assign to variables

    my $header       = &TWiki::Func::extractNameValuePair( $attributes, "header" );
    my $footer       = &TWiki::Func::extractNameValuePair( $attributes, "footer" );
    my $displayon    = &TWiki::Func::extractNameValuePair( $attributes, "displayon" );

    if ( $displayon ) {
        $currentform = &extractFormName;
        if ( $displayon ne $currentform ) { return "" }
    }

    $thread = extractMetaData( "FIELD", "Thread", "value", $web, $topic );
    $tempjoinerval = ",";
    $thread =~ s/^$tempjoinerval//g;                                                 # remove comma from start and end of Thread string
    $thread =~ s/$tempjoinerval$//g;
    @threadvals = split(/,/,$thread);                                                # make an array of thread vals

    $parenttopic = $threadvals[0];                                                   # the first element is the article parent
    $parenttitle = extractMetaData( "FIELD", "Title", "value", $web, $parenttopic ); # get the article title
    
    $return = &TWiki::Func::renderText( "[[$parenttopic][$parenttitle]]" );          # make the link to the parent article
    $return = $header . $return . $footer;

}

sub renderCommentCount {
 ### UNDER DEV: not really working yet ###
    my $tempVal;
    my $count;
    my ( $attributes ) = @_;
    my $alttopic     = &TWiki::Func::extractNameValuePair( $attributes, "topic" );
    my $searchval = "FIELD.*name=\"Thread\".*value=\".*$alttopic.*\";FORM.*name=\"BoardResponseForm\"";
    my $cmdString = "/home/evem/public_html/bin/andgrep -l -i '$searchval' *.txt";
    
    # The following is copied straight from search.pm
    my $sDir = "/home/evem/data/$web";
    my @topicList = ();
    if( $theSearchVal ) {
        # do grep search
        chdir( "$sDir" );
        $tempVal = `$cmdString`;
        @topicList = split( /\n/, $tempVal );
    }

    $count = @topicList;
    return $count;

}

# ========================= #############################
# = /Custom EvEm Bit ====== #############################
# ========================= #############################

# =========================
sub commonTagsHandler
{
### my ( $text, $topic, $web ) = @_;   # do not uncomment, use $_[0], $_[1]... instead

    &TWiki::Func::writeDebug( "- EvEm::commonTagsHandler( $_[2].$_[1] )" ) if $debug;

    # This is the place to define customized tags and variables
    # Called by sub handleCommonTags, after %INCLUDE:"..."%

    # do custom extension rule, like for example:
    # $_[0] =~ s/%XYZ%/&handleXyz()/geo;
    # $_[0] =~ s/%XYZ{(.*?)}%/&handleXyz($1)/geo;

    $_[0] =~ s/%TITLE%/&renderTitle()/geo;
}

# =========================
sub startRenderingHandler
{
### my ( $text, $web ) = @_;   # do not uncomment, use $_[0], $_[1] instead

    &TWiki::Func::writeDebug( "- EvEm::startRenderingHandler( $_[1].$topic )" ) if $debug;

    # This handler is called by getRenderedVersion just before the line loop

    # do custom extension rule, like for example:
    # $_[0] =~ s/old/new/go;
}

# =========================
sub DISABLE_outsidePREHandler
{
### my ( $text ) = @_;   # do not uncomment, use $_[0] instead

#   &TWiki::Func::writeDebug( "- EvEmPlugin::outsidePREHandler( $web.$topic )" ) if $debug;

    # This handler is called by getRenderedVersion, in loop outside of <PRE> tag.
    # This is the place to define customized rendering rules.
    # Note: This is an expensive function to comment out.
    # Consider startRenderingHandler instead
}

# =========================
sub DISABLE_insidePREHandler
{
### my ( $text ) = @_;   # do not uncomment, use $_[0] instead

#   &TWiki::Func::writeDebug( "- EvEmPlugin::insidePREHandler( $web.$topic )" ) if $debug;

    # This handler is called by getRenderedVersion, in loop inside of <PRE> tag.
    # This is the place to define customized rendering rules.
    # Note: This is an expensive function to comment out.
    # Consider startRenderingHandler instead
}

# =========================
sub endRenderingHandler
{
### my ( $text ) = @_;   # do not uncomment, use $_[0] instead

    &TWiki::Func::writeDebug( "- EvEmPlugin::endRenderingHandler( $web.$topic )" ) if $debug;

    # This handler is called by getRenderedVersion just after the line loop

    $_[0] =~ s/%METAFIELD{(.*?)}%/&renderMetaField($1)/geo;
    $_[0] =~ s/%SEARCHLINKS%/&handleKeywords()/geo;
    $_[0] =~ s/%SEARCHLINKS{(.*?)}%/&handleKeywords($1)/geo;
    $_[0] =~ s/\$searchlinks\((.*?)\)/&handleKeywords($1)/geo;
    $_[0] =~ s/%TITLE%/&renderTitle()/geo;
    $_[0] =~ s/%SECTION%/&renderKeywordSection()/geo;
    $_[0] =~ s/%SECTION{(.*?)}%/&renderKeywordSection($1)/geo;
    $_[0] =~ s/%KEYWORDLINKS%/&renderSimilarKeywordLinks()/geo;
    $_[0] =~ s/%KEYWORDLINKS{(.*?)}%/&renderSimilarKeywordLinks($1)/geo;
    $_[0] =~ s/\$depth\((.*?)\)/&renderThread($1)/geo;
    $_[0] =~ s/%FORM{(.*?)}%/&renderForm($1)/geo;
    $_[0] =~ s/%FORMDATA{(.*?)}%/&renderFormData($1)/geo;
    $_[0] =~ s/\$image\((.*?)\)/&renderImage($1)/geo;
    $_[0] =~ s/%IMAGE{(.*?)}%/&renderImage($1)/geo;
    $_[0] =~ s/\$username\((.*?)\)/&renderAuthorUserName($1)/geo;
    $_[0] =~ s/%DOCID%/&renderDocumentID()/geo;
    $_[0] =~ s/%RESPONSEPARENTLINK{(.*?)}%/&renderResponseParentLink($1)/geo;
    $_[0] =~ s/\$commentcount\((.*?)\)/&renderCommentCount($1)/geo;
    $_[0] =~ s/%ENV{(.*?)}%/&handleEnvVariable($1)/geo;

}

# =========================

1;
